local function GetSortArrow(ascending)
    local s = "|TInterface\\Addons\\Sorted\\Textures\\Sort-Arrow"
    if Sorted_GetSetting("fontOutline") > 0 then
        s = s.."-Outline"
        if Sorted_GetSetting("fontOutline") > 1 then
            s = s..":0:0:-5"
        else
            s = s..":0:0:-2"
        end
    else
        s = s..":0:0:-1"
    end
    if ascending then
        s = s..":-1:16:16:0:16:0:16|t"
    else
        s = s..":0:16:16:0:16:16:0|t"
    end
    return s
end

-- Sort functions
local function SortByName(slot1, slot2) return slot1.name < slot2.name end
local function SortByQuantity(slot1, slot2) return slot1.count < slot2.count end
local function SortByLevel(slot1, slot2) if not slot1.effectiveILvl then return false elseif not slot2.effectiveILvl then return true end return slot1.effectiveILvl > slot2.effectiveILvl end
local function SortByValue(slot1, slot2) return slot1.value * slot1.count < slot2.value * slot2.count end
local function SortByRarity(slot1, slot2) return slot1.rarity > slot2.rarity end
local function SortByNameThenQuantity(slot1, slot2)
    if slot1.name == slot2.name then
        if slot1.count == slot2.count then
            return slot1.bag * 36 + slot1.slot < slot2.bag * 36 + slot2.slot
        end
        return SortByQuantity(slot2, slot1)
    end
    return SortByName(slot1, slot2)
end
local function SortByRarityThenLevel(slot1, slot2)
    if slot1.rarity == slot2.rarity then
        if slot1.effectiveILvl == slot2.effectiveILvl then
            return SortByNameThenQuantity(slot1, slot2)
        end
        return SortByLevel(slot1, slot2)
    else
        return SortByRarity(slot1, slot2)
    end
end
local function SortByLevelThenRarity(slot1, slot2)
    if slot1.effectiveILvl == slot2.effectiveILvl then
        if slot1.rarity == slot2.rarity then
            return SortByNameThenQuantity(slot1, slot2)
        end
        return SortByRarity(slot1, slot2)
    else
        return SortByLevel(slot1, slot2)
    end
end

local function Sorted_SlotSort(slot1, slot2)
    if not slot1 then return false
    elseif not slot2 then return true end
    if slot1.empty then return false
    elseif slot2.empty then return true end
    if slot1.filtered and not slot2.filtered then return false
    elseif slot2.filtered and not slot1.filtered then return true end
    if not slot1.itemID then return false
    elseif not slot2.itemID then return true end

    if Sorted_GetSetting("favoritesOnTop") == 2 then
        local slot1Fav, slot2Fav = Sorted_GetFavorited(slot1), Sorted_GetFavorited(slot2)
        if slot1Fav ~= slot2Fav then
            if slot1Fav == nil then return false
            elseif slot2Fav == nil then return true
            else return slot1Fav < slot2Fav end
        end
    end

    local sortButton = Sorted_sortButtons[Sorted_selectedSort]

    if Sorted_sortVariant < 3 or #sortButton.sortMethods == 1 then
        if Sorted_sortVariant % 2 == 1 then
            return sortButton.sortMethods[1].func(slot1, slot2)
        else
            return sortButton.sortMethods[1].func(slot2, slot1)
        end
    else
        if Sorted_sortVariant % 2 == 1 then
            return sortButton.sortMethods[2].func(slot1, slot2)
        else
            return sortButton.sortMethods[2].func(slot2, slot1)
        end
    end
end

function SortedItemList_SortItems(self)
    table.sort(self.itemButtons, Sorted_SlotSort)
end


-- SORT BUTTONS

-- These will always refer to the sortButtonle position of the sortButton in Sorted_sortButtons, in case the order is changed in the future
SORTED_SORTBUTTON_FAVORITES = 1
SORTED_SORTBUTTON_QUANTITY = 2
SORTED_SORTBUTTON_NAME = 3
SORTED_SORTBUTTON_REQUIRED_LEVEL = 4
SORTED_SORTBUTTON_LEVEL = 5
SORTED_SORTBUTTON_ADDED = 6
SORTED_SORTBUTTON_TYPE = 7
SORTED_SORTBUTTON_EXPANSION = 8
SORTED_SORTBUTTON_BIND = 9
SORTED_SORTBUTTON_TRANSMOG = 10
SORTED_SORTBUTTON_VALUE = 11

-- Saved between sessions. Uses unique IDs
Sorted_enabledTabs = {
    [1] = true,  -- Favorites
    [2] = true,  -- Quantity
    [3] = true,  -- Name
    [4] = true,  -- Item Level
    [5] = true,  -- Value
    [6] = false, -- Time Added
    [7] = false, -- Binding
    [8] = false, -- Req Level
    [9] = false, -- Type
}

-- Ordered left to right
Sorted_sortButtons = {
    {
        ["name"] = "Favourites",
        ["button"] = SortedFrameFavoritesSortButton,
        ["width"] = 20,
        ["itemButtonKeys"] = {"favoriteButton"},
        ["uniqueID"] = 1 -- This will never change in the future, even if other sortButtons are added.
    },
    {
        ["name"] = "Quantity",
        ["button"] = SortedFrameQuantitySortButton,
        ["width"] = 42,
        ["sortMethods"] = {
            {
                ["title"] = "#",
                ["func"] = function(slot1, slot2)
                    if slot1.count == slot2.count then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    return SortByQuantity(slot2, slot1)
                end
            }
        },
        ["align"] = "RIGHT",
        ["itemButtonKeys"] = {"quantityString"},
        ["uniqueID"] = 2
    },
    {
        ["name"] = "Name",
        ["button"] = SortedFrameNameSortButton,
        ["width"] = 0,
        ["sortMethods"] = {
            {
                ["title"] = "Rarity",
                ["func"] = function(slot1, slot2) 
                    return SortByRarityThenLevel(slot1, slot2)
                end
            },
            {
                ["title"] = "Name",
                ["func"] = function(slot1, slot2) 
                    return SortByNameThenQuantity(slot1, slot2)
                end,
                ["inverse"] = true
            }
        },
        ["align"] = "LEFT",
        ["itemButtonKeys"] = {"nameString"},
        ["uniqueID"] = 3
    },
    {
        ["name"] = "Required Level",
        ["button"] = SortedFrameReqLevelSortButton,
        ["width"] = 46,
        ["sortMethods"] = {
            {
                ["title"] = "Req",
                ["func"] = function(slot1, slot2)
                    if slot1.minLevel == slot2.minLevel or (slot1.minLevel < 2 and slot2.minLevel < 2) then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    return slot1.minLevel > slot2.minLevel
                end
            }
        },
        ["align"] = "LEFT",
        ["itemButtonKeys"] = {"minLevelString"},
        ["uniqueID"] = 8
    },
    {
        ["name"] = "Item Level",
        ["button"] = SortedFrameLevelSortButton,
        ["width"] = 46,
        ["sortMethods"] = {
            {
                ["title"] = "Lvl",
                ["func"] = function(slot1, slot2) 
                    return SortByLevelThenRarity(slot1, slot2)
                end
            }
        },
        ["align"] = "LEFT",
        ["itemButtonKeys"] = {"levelString", "UpgradeIcon"},
        ["uniqueID"] = 4
    },
    {
        ["name"] = "Time Added",
        ["button"] = SortedFrameAddedSortButton,
        ["width"] = 64,
        ["sortMethods"] = {
            {
                ["title"] = "Age",
                ["func"] = function(slot1, slot2)
                    if not slot1.timeAdded then return false end
                    if not slot2.timeAdded then return true end
                    if slot1.timeAdded == slot2.timeAdded then
                        return SortByNameThenQuantity(slot1, slot2)
                    end
                    return slot1.timeAdded > slot2.timeAdded
                end
            }
        },
        ["align"] = "RIGHT",
        ["itemButtonKeys"] = {"timeAddedString"},
        ["uniqueID"] = 6
    },
    {
        ["name"] = "Type",
        ["button"] = SortedFrameTypeSortButton,
        ["width"] = 80,
        ["sortMethods"] = {
            {
                ["title"] = "Type",
                ["func"] = function(slot1, slot2)
                    if slot1.type == slot2.type then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    return slot1.type < slot2.type
                end
            }
        },
        ["align"] = "LEFT",
        ["itemButtonKeys"] = {"typeString"},
        ["uniqueID"] = 9
    },
    {
        ["name"] = "Expansion",
        ["button"] = SortedFrameExpacSortButton,
        ["width"] = 64,
        ["sortMethods"] = {
            {
                ["title"] = "Expac",
                ["func"] = function(slot1, slot2)
                    if slot1.expacID == slot2.expacID then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    if not slot1.expacID then return false end
                    if not slot2.expacID then return true end
                    return slot1.expacID > slot2.expacID
                end
            }
        },
        ["align"] = "LEFT",
        ["itemButtonKeys"] = {"expacString"},
        ["uniqueID"] = 10
    },
    {
        ["name"] = "Binding",
        ["button"] = SortedFrameBindSortButton,
        ["width"] = 48,
        ["sortMethods"] = {
            {
                ["title"] = "Bind",
                ["func"] = function(slot1, slot2)
                    if slot1.bindType == slot2.bindType then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    if slot1.bindType == 0 then return false end
                    if slot2.bindType == 0 then return true end
                    return slot1.bindType > slot2.bindType
                end
            }
        },
        ["align"] = "RIGHT",
        ["itemButtonKeys"] = {"bindString"},
        ["uniqueID"] = 7
    },
    {
        ["name"] = "Sell Price",
        ["button"] = SortedFrameValueSortButton,
        ["width"] = 56,
        ["sortMethods"] = {
            {
                ["title"] = "Value",
                ["func"] = function(slot1, slot2) 
                    if slot1.hasNoValue and slot2.hasNoValue then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    if slot1.hasNoValue then return false end
                    if slot2.hasNoValue then return true end
                    if slot1.value * slot1.count == slot2.value * slot2.count then
                        return SortByNameThenQuantity(slot1, slot2)
                    end
                    return slot1.value * slot1.count > slot2.value * slot2.count
                end
            }
        },
        ["align"] = "RIGHT",
        ["itemButtonKeys"] = {"valueString"},
        ["uniqueID"] = 5
    },
    {
        ["name"] = "Transmog Collected",
        ["button"] = SortedFrameTransmogSortButton,
        ["width"] = 64,
        ["sortMethods"] = {
            {
                ["title"] = "Mog",
                ["func"] = function(slot1, slot2)
                    if slot1.transmogValue == slot2.transmogValue then
                        return SortByRarityThenLevel(slot1, slot2)
                    end
                    return slot1.transmogValue > slot2.transmogValue
                end
            }
        },
        ["align"] = "RIGHT",
        ["itemButtonKeys"] = {"transmogString"},
        ["uniqueID"] = 11,
        ["disabled"] = true
    }
}
SORTED_DEFAULT_SORT = 3

function Sorted_UpdateSortButtons()
    -- Left sortButtons
    local leftPosX, rightPosX = 0, 0
    local indexReached = 0
    for i, v in ipairs(Sorted_sortButtons) do
        if not v.disabled then
            indexReached = i
            if v.width == 0 then
                break -- Stop at item name column
            end
            if Sorted_enabledTabs[v.uniqueID] then
                v.button:SetWidth(v.width)
                v.button:SetPoint("LEFT", leftPosX, 0)
                if v.itemButtonKeys then
                    for _,itemList in pairs(Sorted_itemLists) do
                        for _,slot in pairs(itemList.itemButtons) do
                            for _,itemButtonKey in pairs(v.itemButtonKeys) do
                                if itemButtonKey ~= "favoriteButton" then
                                    if v.align == "LEFT" then
                                        slot[itemButtonKey]:SetPoint("LEFT", slot, "LEFT", 10 + leftPosX, 0)
                                    elseif v.align == "RIGHT" then
                                        slot[itemButtonKey]:SetPoint("RIGHT", slot, "LEFT", -2 + leftPosX + v.width, 0)
                                    end
                                    slot[itemButtonKey]:SetWidth(v.width)
                                end
                                if slot[itemButtonKey].SetFrameLevel then
                                    slot[itemButtonKey]:SetFrameLevel(580)
                                end
                            end
                        end
                    end
                end
                leftPosX = leftPosX + v.width
                if v.columnSeparators then
                    for k,colSep in pairs(v.columnSeparators) do   
                        colSep:SetPoint("LEFT", leftPosX - 2, 0)
                    end
                end
            end
        end
    end
    -- Right sortButtons
    for i = #Sorted_sortButtons, indexReached + 1, -1 do
        local v = Sorted_sortButtons[i]
        if not v.disabled then
            if Sorted_enabledTabs[v.uniqueID] then
                v.button:SetWidth(v.width)
                v.button:SetPoint("RIGHT", rightPosX - 4, 0)
                if v.itemButtonKeys then
                    for _,itemList in pairs(Sorted_itemLists) do
                        for _,slot in pairs(itemList.itemButtons) do
                            for _,itemButtonKey in pairs(v.itemButtonKeys) do
                                if v.align == "LEFT" then
                                    slot[itemButtonKey]:SetPoint("LEFT", slot, "RIGHT", 2 + rightPosX - v.width, 0)
                                elseif v.align == "RIGHT" then
                                    slot[itemButtonKey]:SetPoint("RIGHT", slot, "RIGHT", -8 + rightPosX, 0)
                                end
                                if slot[itemButtonKey].SetFrameLevel then
                                    slot[itemButtonKey]:SetFrameLevel(580)
                                end
                            end
                        end
                    end
                end
                if v.columnSeparators then
                    for k,colSep in pairs(v.columnSeparators) do
                        if rightPosX == 0 then
                            colSep:SetPoint("LEFT", colSep.relScrollBar, "LEFT", rightPosX - 5, 0)
                        else
                            colSep:SetPoint("LEFT", colSep.relScrollBar, "LEFT", rightPosX - 4, 0)
                        end
                    end
                end
                rightPosX = rightPosX - v.width
            end
        end
    end
    -- Item name sortButton
    Sorted_sortButtons[indexReached].button:SetPoint("LEFT", leftPosX, 0)
    Sorted_sortButtons[indexReached].button:SetPoint("RIGHT", rightPosX - 4, 0)
    for k,itemList in pairs(Sorted_itemLists) do
        for l,slot in pairs(itemList.itemButtons) do
            if leftPosX == 0 then
                slot.iconFrame:SetPoint("LEFT", slot, "LEFT", 6, 0)
            else
                slot.iconFrame:SetPoint("LEFT", slot, "LEFT", 2 + leftPosX, 0)
            end
            slot.nameFrame:SetPoint("LEFT", slot.iconFrame, "RIGHT", 4, 0)
            slot.nameFrame:SetPoint("RIGHT", slot, "RIGHT", rightPosX, 0)
        end
    end
    if Sorted_sortButtons[indexReached].columnSeparators then
        for k,colSep in pairs(Sorted_sortButtons[indexReached].columnSeparators) do
            colSep:SetPoint("LEFT", colSep.relScrollBar, "LEFT", rightPosX - 4, 0)
        end
    end

    SortedFrame_UpdateMinSize()
end

function Sorted_EnableSortButton(id)
    Sorted_sortButtons[id].button:Show()
    if Sorted_sortButtons[id].columnSeparators then
        for k,v in pairs(Sorted_sortButtons[id].columnSeparators) do   
            v:Show()
        end
    end
    if Sorted_sortButtons[id].itemButtonKeys then
        for _,itemList in pairs(Sorted_itemLists) do
            for _,slot in pairs(itemList.itemButtons) do
                for _,itemButtonKey in pairs(Sorted_sortButtons[id].itemButtonKeys) do
                    slot[itemButtonKey]:Show()
                    if slot[itemButtonKey].SetFrameLevel then
                        slot[itemButtonKey]:SetFrameLevel(580)
                    end
                end
            end
        end
    end
    Sorted_enabledTabs[Sorted_sortButtons[id].uniqueID] = true
    Sorted_UpdateSortButtons()
    Sorted_UpdateTimeItemsAdded()
    Sorted_UpdateItemButtons()
    Sorted_FilterItems()
end
function Sorted_DisableSortButton(id)
    if Sorted_selectedSort == id then
        Sorted_selectedSort = SORTED_DEFAULT_SORT
        Sorted_sortVariant = 1
        Sorted_sortButtons[SORTED_DEFAULT_SORT].button.text:SetText("Rarity")
        Sorted_SortItems()
    end
    if Sorted_sortButtons[id].columnSeparators then
        for k,v in pairs(Sorted_sortButtons[id].columnSeparators) do
            v:Hide()
        end
    end
    if Sorted_sortButtons[id].itemButtonKeys then
        for _,itemList in pairs(Sorted_itemLists) do
            for _,slot in pairs(itemList.itemButtons) do
                for _,itemButtonKey in pairs(Sorted_sortButtons[id].itemButtonKeys) do
                    slot[itemButtonKey]:Hide()
                end
            end
        end
    end
    Sorted_sortButtons[id].button:Hide()
    Sorted_enabledTabs[Sorted_sortButtons[id].uniqueID] = false
    Sorted_UpdateSortButtons()
    Sorted_UpdateItemButtons()
    Sorted_FilterItems()
end
function Sorted_ToggleSortButton(id)
    if Sorted_enabledTabs[Sorted_sortButtons[id].uniqueID] then
        Sorted_DisableSortButton(id)
    else
        Sorted_EnableSortButton(id)
    end
end

function SortedSortButton_OnClick(self, button)
    PlaySound(SOUNDKIT.IG_CHAT_SCROLL_UP)
    if SortedFrameTabsMenu:IsShown() then
        SortedFrameTabsMenu:Hide()
    else
        if button == "LeftButton" then
            if Sorted_selectedSort == self:GetID() then
                Sorted_sortVariant = Sorted_sortVariant + 1
                if Sorted_sortVariant > #Sorted_sortButtons[Sorted_selectedSort].sortMethods * 2 then 
                    Sorted_sortVariant = 1 
                end
            else
                Sorted_selectedSort = self:GetID()
                Sorted_sortVariant = 1
            end
            Sorted_sortButtons[SORTED_SORTBUTTON_NAME].button.text:SetText("Name") 
            SortedSortButtons_Update()
            Sorted_SortItems()
            Sorted_ScrollToTop()
        else
            local x,y = GetCursorPosition()
            x,y = x/UIParent:GetEffectiveScale(), y/UIParent:GetEffectiveScale()
            SortedFrameTabsMenu:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x * (1 / SortedFrameTabsMenu:GetScale()), y * (1 / SortedFrameTabsMenu:GetScale()))
            SortedFrameTabsMenu:Show()
        end
    end
end

function SortedSortButtons_Update()
    for i, button in ipairs(Sorted_sortButtons) do
        if i ~= SORTED_SORTBUTTON_FAVORITES then
            local s
            if i == Sorted_selectedSort then
                s = button.sortMethods[math.ceil(Sorted_sortVariant / 2)].title
            else
                s = button.sortMethods[#button.sortMethods].title
            end
            if i == Sorted_selectedSort then
                local ascending
                if button.sortMethods[math.ceil(Sorted_sortVariant / 2)].inverse then
                    ascending = ((Sorted_sortVariant % 2) == 0)
                else
                    ascending = ((Sorted_sortVariant % 2) == 1)
                end
                s = s..GetSortArrow(ascending)
            end
            button.button.text:SetText(s)
        end
    end
end

function SortedSortButtons_OnLoad(self)
    Sorted_selectedSort = SORTED_DEFAULT_SORT
    Sorted_sortVariant = 1
    for i,v in ipairs(Sorted_sortButtons) do
        if i ~= 1 then -- Ignore favorites tab
            local b = CreateFrame("Button", self:GetName()..v.name, self, "SortedColumnHeaderTemplate")
            b:SetID(i)
            b:RegisterForClicks("LeftButtonDown", "RightButtonDown")
            b:SetScript("OnClick", SortedSortButton_OnClick)

            v.button = b

            if v.sortMethods then
                b.text:SetText(v.sortMethods[1].title) 
            end
        end
    end
    Sorted_sortButtons[1].button = SortedFrameFavoritesSortButton
    SortedSortButtons_Update()
    
    -- Tab dropdown menu
    local tabMenu = CreateFrame("Frame", "SortedFrameTabsMenu", UIParent, "TooltipBorderedFrameTemplate")
    if Sorted_IsDemo() then
        tabMenu:SetFrameStrata("FULLSCREEN_DIALOG")
    else
        tabMenu:SetFrameStrata("DIALOG")
    end
    tabMenu.checkButtons = {}
    function tabMenu.UpdateWidth(self)
        local width = 0
        for k,v in pairs(self.checkButtons) do
            if v.text:GetWidth() > width then
                width = v.text:GetWidth()
            end
        end
        self:SetWidth(36 + width)
    end
    local y = 0
    for i,v in ipairs(Sorted_sortButtons) do
        if i ~= SORTED_SORTBUTTON_NAME and not (Sorted_IsClassic() and i == SORTED_SORTBUTTON_EXPANSION) and not Sorted_sortButtons[i].disabled then
            local cb = CreateFrame("CheckButton", "SortedFrameTabsMenu"..v.name, tabMenu, "SortedCheckButtonTemplate")
            cb.text:ClearAllPoints()
            cb.text:SetPoint("LEFT", cb, "RIGHT")
            cb.text:SetText(v.name)
            cb:SetPoint("TOPLEFT", 3, -4 - y)
            cb.clickable:SetPoint("RIGHT", tabMenu)
            cb.id = i
            cb:SetScript("OnClick", function(self, button)
                if button == "LeftButton" then
                    Sorted_ToggleSortButton(self.id)
                    self:SetChecked(Sorted_enabledTabs[Sorted_sortButtons[self.id].uniqueID])
                end
            end)
            tabMenu.checkButtons[i] = cb
            y = y + cb.clickable:GetHeight()
            cb:RegisterEvent("PLAYER_ENTERING_WORLD")
            cb:HookScript("OnEvent", function(self, event)
                if event == "PLAYER_ENTERING_WORLD" then
                    self:SetChecked(Sorted_enabledTabs[Sorted_sortButtons[self.id].uniqueID])
                end
            end)
        end
    end
    tabMenu:SetHeight(y + 11)
    tabMenu:UpdateWidth()
    tabMenu:Hide()
    if not Sorted_IsClassic() then
        tabMenu:RegisterEvent("GLOBAL_MOUSE_DOWN")
    end
    tabMenu:SetScript("OnEvent", function(self)
        if self:IsShown() and not self:IsMouseOver() then 
            self:Hide() 
        end
    end)
    tabMenu:SetScript("OnShow", function(self)
        self:UpdateWidth()
    end)
end